package explorviz.live_trace_processing.filter.reduction;

import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import explorviz.live_trace_processing.Constants;
import explorviz.live_trace_processing.filter.AbstractFilter;
import explorviz.live_trace_processing.filter.IPipeReceiver;
import explorviz.live_trace_processing.reader.TimeProvider;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.event.AbstractBeforeEventRecord;
import explorviz.live_trace_processing.record.event.AbstractBeforeOperationEventRecord;
import explorviz.live_trace_processing.record.event.AbstractEventRecord;
import explorviz.live_trace_processing.record.misc.TerminateRecord;
import explorviz.live_trace_processing.record.misc.TimedPeriodRecord;
import explorviz.live_trace_processing.record.trace.Trace;
import explorviz.live_trace_processing.record.trace.TraceComperator;

public class TracesSummarizationFilter extends AbstractFilter implements ITraceReduction {
	private final long maxCollectionDuration;

	private final Map<Trace, TracesSummarizationBuffer> trace2buffer = new TreeMap<Trace, TracesSummarizationBuffer>(
			new TraceComperator());

	public TracesSummarizationFilter(final long maxCollectionDuration,
			final IPipeReceiver sinkReceiver) {
		super(sinkReceiver, Constants.TRACE_SUMMARIZATION_DISRUPTOR_SIZE,
				Constants.TRACE_SUMMARIZATION_OUTPUT_BUFFER_SIZE, "Reduced traces/sec", 1000);

		this.maxCollectionDuration = maxCollectionDuration;
	}

	@Override
	public void processRecord(final IRecord record) {
		if (record instanceof Trace) {
			final Trace trace = (Trace) record;

			if (trace.isValid()) {
				if (!trace.containsRemoteRecord()) {
					insertIntoBuffer(trace);
				} else {
					// trace with remote records cant be reduced
					makeTraceElementsAccumulator(trace);
					deliver(trace);
				}
			} else {
				makeTraceElementsAccumulator(trace);
				deliver(trace);
			}
		} else if (record instanceof TimedPeriodRecord) {
			processTimeoutQueue(TimeProvider.getCurrentTimestamp());
			periodicFlush(record);
			// deliver(record);
		} else if (record instanceof TerminateRecord) {
			terminate();
			deliver(record);
		} else {
			deliver(record);
		}
	}

	private void insertIntoBuffer(final Trace trace) {
		TracesSummarizationBuffer traceAggregationBuffer = trace2buffer.get(trace);
		if (traceAggregationBuffer == null) {
			makeTraceElementsAccumulator(trace);

			traceAggregationBuffer = new TracesSummarizationBuffer(
					TimeProvider.getCurrentTimestamp());
			trace2buffer.put(trace, traceAggregationBuffer);
		}
		traceAggregationBuffer.insertTrace(trace);
	}

	public void makeTraceElementsAccumulator(final Trace trace) {
		for (int i = 0; i < trace.getTraceEvents().size(); i++) {
			final AbstractEventRecord event = trace.getTraceEvents().get(i);
			if (event instanceof AbstractBeforeEventRecord) {
				final AbstractBeforeEventRecord abstractBeforeEventRecord = (AbstractBeforeEventRecord) event;

				if (abstractBeforeEventRecord instanceof AbstractBeforeOperationEventRecord) {
					final AbstractBeforeOperationEventRecord abstractBeforeOperationEventRecord = (AbstractBeforeOperationEventRecord) abstractBeforeEventRecord;

					abstractBeforeOperationEventRecord.getRuntimeStatisticInformation()
							.makeAccumulator(abstractBeforeOperationEventRecord.getObjectId());
				} else {
					abstractBeforeEventRecord.getRuntimeStatisticInformation().makeAccumulator(0);

				}
			}
		}
	}

	private void processTimeoutQueue(final long timestamp) {
		final long bufferTimeout = timestamp - maxCollectionDuration;
		final Iterator<Entry<Trace, TracesSummarizationBuffer>> iter = trace2buffer.entrySet()
				.iterator();

		while (iter.hasNext()) {
			final Entry<Trace, TracesSummarizationBuffer> traceBufferEntry = iter.next();

			if (traceBufferEntry.getValue().getBufferCreatedTimestamp() <= bufferTimeout) {
				final Trace aggregatedTrace = traceBufferEntry.getValue().getAggregatedTrace();

				deliver(aggregatedTrace);
				iter.remove();
			}
		}
	}

	private void terminate() {
		for (final TracesSummarizationBuffer traceBuffer : trace2buffer.values()) {
			deliver(traceBuffer.getAggregatedTrace());
		}
		trace2buffer.clear();
	}
}
